#!/usr/bin/python
# -*- coding: utf-8 -*-
### BEGIN LICENSE
# Copyright (C) 2010 Mads Chr. Olesen <mchro@cs.aau.dk>
#This program is free software: you can redistribute it and/or modify it 
#under the terms of the GNU General Public License version 3, as published 
#by the Free Software Foundation.
#
#This program is distributed in the hope that it will be useful, but 
#WITHOUT ANY WARRANTY; without even the implied warranties of 
#MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
#PURPOSE.  See the GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License along 
#with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import pyuppaal
from pyuppaal.ulp import lexer, parser
import util
import copy
import re

class SimplifyModel:

    def __init__(self, nta, constant_overrides=dict()):
        self.nta = nta
        self.constant_overrides = constant_overrides
        self.parse_and_calculate_constants()

    def parse_and_calculate_constants(self):
        self.lex = lexer.lexer
        self.pars = parser.Parser(self.nta.declaration, self.lex)
        self.declvisitor = parser.DeclVisitor(self.pars)

        self.constants = {}
        #constant overrides
        for (cident, cval) in self.constant_overrides.iteritems():
            #XXX assumes expression is already python
            self.constants[cident] = eval(cval, {}, self.constants)
        #evaluate constants to values
        for (cident, cval) in self.declvisitor.constants.iteritems():
            if not cident in self.constants:
                self.constants[cident] = eval(util.expression_to_python(cval), {}, self.constants)

    def simplify(self):
        self.expand_system_declaration()
        self.expand_selects()
        return self.nta

    def expand_selects(self):
        """Expand transitions with select based on the possible values for the 
        select variables."""
        for temp in self.nta.templates:
            new_transitions = []
            for trans in temp.transitions:
                if trans.select.value:
                    for var in [v.strip() for v in trans.select.value.split(',')]:
                        #TODO: not correct for more than one variable
                        (ident, typename) = var.split(':')
                        ident = ident.strip()
                        typename = typename.strip()

                        lower, upper = self.bounds_for_type(typename)

                        for x in range(lower, upper+1):
                            new_trans = copy.copy(trans)
                            new_trans.select.value = ''
                            self.inline_var_in_trans(new_trans, ident, x)
                            new_transitions.append(new_trans)
            temp.transitions = [t for t in temp.transitions if not t.select.value] + new_transitions
                    

    def bounds_for_type(self, typename):
        type = self.pars.getType(typename)
        assert type.children[0].type == 'TypeInt' #TODO other types
        lowerexpr = util.expression_to_python(type.children[0].children[0].children[0])
        upperexpr = util.expression_to_python(type.children[0].children[1].children[0])

        lower = eval(lowerexpr, {}, self.constants)
        upper = eval(upperexpr, {}, self.constants)

        return (lower, upper)


    def expand_system_declaration(self):
        """Expand the model according to the system declaration:
        * For each automaton in the system create a unique template, inlining parameters
        * TODO: Promote local variables to uniquely named global vars.
        """
        nta = self.nta
        ret = pyuppaal.NTA(declaration=nta.declaration)
        systemsline = [l for l in nta.system.split('\n') if l.strip().startswith('system')][-1]
        systemsline.strip()
        systemsline = systemsline[len('systems ')-1:-1]



        
        for s in [s.strip() for s in systemsline.split(',')]:
            temp = [t for t in nta.templates if t.name == s][0]
            #case with no pars
            if temp.parameter.strip() == '':
                ret.templates += [temp]
            #case where we instantiate based on possible parameter values
            else:
                comb_temps = [copy.deepcopy(temp)]
                comb_temps[0].parameter = ''
                for par in [p.strip() for p in temp.parameter.split(',')]:
                    (typename, ident) = par.split(' ')
                    lower, upper = self.bounds_for_type(typename)
                    
                    new_temps = []
                    for oldtemp in comb_temps:
                        for x in range(lower, upper+1):
                            new_temp = self.inline_var_in_temp(oldtemp, ident, x)
                            new_temp.name += '_' + str(x)
                            new_temps += [new_temp]
                    comb_temps = new_temps
                ret.templates += comb_temps
            #TODO other cases
                            

        ret.system = "system " + ", ".join([t.name for t in ret.templates]) + ";"

        self.nta = ret

    def inline_var_in_temp(self, temp, ident, val):
        """Returns a new template, with ident inlined to the given val."""
        regex = re.compile('(?<!\w)%s(?!\w)' % ident)
        toret = copy.deepcopy(temp)
        for trans in toret.transitions:
            self.inline_var_in_trans(trans, ident, val)
        #TODO: inline var in invariants
        return toret

    def inline_var_in_trans(self, trans, ident, val):
        """Inlines ident to the given val, in trans."""
        regex = re.compile('(?<!\w)%s(?!\w)' % ident)
        trans.select.value = re.sub(regex, str(val), trans.select.value)
        trans.guard.value = re.sub(regex, str(val), trans.guard.value)
        trans.assignment.value = re.sub(regex, str(val), trans.assignment.value)
        trans.synchronisation.value = re.sub(regex, str(val), trans.synchronisation.value)

if __name__ == '__main__':
    import sys
    file = open(sys.argv[1])
    nta = pyuppaal.from_xml(file)

    out_nta = SimplifyModel(nta).simplify()

    outfile = open(sys.argv[2], 'w')
    outfile.write(out_nta.to_xml())
