#!/usr/bin/python
import sys
import os
import unittest
import logging

import pyuppaal
from opaal.model_parsers.pyuppaal import SuccessorGenerator, GoalChecker
from opaal.passed_waiting_list import SimplePWList, LatticeCoverPWList, LatticeJoinPWList
from opaal.modelchecking_algorithms.reachability import McReachability
from opaal.lattices.abstract_cache import AbstractCache
from opaal.lattices.bitvector import UnionBitVector, IntersectionBitVector
from opaal.lattices import LatticeVector
from opaal.lattices.integer_lattice import MaxInteger

class TestPYUPPAALLattice(unittest.TestCase):

    def test_lattice(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'master-election-simple.xml'))
        nta = pyuppaal.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 7)

    def test_lattice_cover(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'master-election-simple.xml'))
        nta = pyuppaal.from_xml(file)
        succgen = SuccessorGenerator(nta)

        #Node 0 can become master, simplepwlist
        goalchecker = GoalChecker(succgen, 'E<> Node_0.master')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)

        #Node 0 can become master, cover pwlist
        pwlist = LatticeCoverPWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)

        #Node 0 can become master, joining pwlist
        pwlist = LatticeJoinPWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)



    def test_master_elec_simple2(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'master-election-simple2.xml'))
        nta = pyuppaal.from_xml(file)
        
        #Node 0 can become master, should be true
        succgen = SuccessorGenerator(nta)
        goalchecker = GoalChecker(succgen, 'E<> Node_0.master')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()

        succs = list(succgen.successors(init_state))
        self.assertEqual(len(succs), 2)

        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)
        #self.assertEqual(len(pwlist.passed_waiting)

        #Node 1 can become master, should be false
        goalchecker = GoalChecker(succgen, 'E<> Node_1.master')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertFalse(res)
        self.assertEqual(8, len(pwlist.passed_waiting))

    def test_abstract_cache(self):
        a = AbstractCache()
        b = AbstractCache()

        def assert_empty(a):
            for i in range(a.num_cachelines):
                self.assertEqual(len(a.cachelines[i]), 0)

        assert_empty(a)
        assert_empty(b)
        
        self.assertFalse(a.ishit(42))
        a.access(42)
        self.assertTrue(a.ishit(42))
        self.assertEqual(a.cachelines[0], set([42]))
        for i in range(1, a.num_cachelines):
            self.assertEqual(len(a.cachelines[i]), 0)

        c = a.must_join(b)
        assert_empty(c)

        b.access(43)
        c = a.must_join(b)
        assert_empty(c)


        b = AbstractCache()
        b.access(42)
        b.access(43)
        self.assertTrue(b.ishit(42))
        self.assertTrue(b.ishit(43))
        #b = [42, 43], a = [42]
        self.assertEqual(b.cachelines[0], set([43]))
        self.assertEqual(b.cachelines[1], set([42]))
        for i in range(2, b.num_cachelines):
            self.assertEqual(len(b.cachelines[i]), 0)


        #check cached access does not change
        b.access(43)
        self.assertEqual(b.cachelines[0], set([43]))
        self.assertEqual(b.cachelines[1], set([42]))
        for i in range(2, b.num_cachelines):
            self.assertEqual(len(b.cachelines[i]), 0)

        #check hit on second line works correct
        b.access(42)
        self.assertEqual(b.cachelines[0], set([42]))
        self.assertEqual(b.cachelines[1], set([43]))
        for i in range(2, b.num_cachelines):
            self.assertEqual(len(b.cachelines[i]), 0)

        c = a.must_join(b)
        self.assertEqual(c.cachelines[0], set([42]))
        for i in range(1, c.num_cachelines):
            self.assertEqual(len(c.cachelines[i]), 0)

        #b = [42, 43], a = [42]
        a.access(43)
        #b = [42, 43], a = [43, 42]
        c = a.must_join(b)
        self.assertEqual(c.cachelines[0], set([]))
        self.assertEqual(c.cachelines[1], set([43, 42]))
        for i in range(2, c.num_cachelines):
            self.assertEqual(len(c.cachelines[i]), 0)

    def test_abstract_cache_ordering(self):
        a = AbstractCache() ; b = AbstractCache()
        c = a.join(b)
        self.assertTrue(a == b)
        self.assertTrue(a == c)
        self.assertTrue(b == c)

        self.assertTrue(a <= b)
        a.access(42) ; b.access(42)
        self.assertTrue(a <= b)

        a = AbstractCache() ; b = AbstractCache()
        a.access(42) ; b.access(43)
        self.assertFalse(a == b)
        self.assertFalse(a <= b)

        a = AbstractCache() ; b = AbstractCache()
        a.access(42) ; b.access(42)
        a.access(43)
        c = a.must_join(b) #c = [ [], [42], ... ]
        self.assertEqual(c.cachelines[0], set([]))
        self.assertEqual(c.cachelines[1], set([42]))
        self.assertTrue(a <= c)
        self.assertFalse(a == c)

        a = AbstractCache() ; b = AbstractCache()
        a.access(42) ; a.access(43)
        b.access(43) ; b.access(42)
        # a = [43, 42]  b = [42, 43]
        self.assertFalse(a <= b)
        self.assertFalse(b <= a)

    def test_lattice_coverpwlist(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_lattice_cache.xml'))
        nta = pyuppaal.from_xml(file)
        
        succgen = SuccessorGenerator(nta)
        goalchecker = GoalChecker(succgen, '')
        pwlist = LatticeCoverPWList()

        state1 = succgen.get_initialstate()
        state2 = state1.copy()

        a = AbstractCache() ; b = AbstractCache()
        state1.lattice_part['cache'] = a ; state2.lattice_part['cache'] = b

        self.assertTrue(state1 <= state2)
        self.assertTrue(state1 == state2)

        b.access(42)
        self.assertFalse(state1 <= state2)
        self.assertFalse(state1 == state2)

        a.access(43)
        self.assertFalse(state1 <= state2)
        self.assertFalse(state1 == state2)

        a.access(42)
        self.assertTrue(state1 <= state2)
        self.assertFalse(state1 == state2)


        a = AbstractCache() ; b = AbstractCache()
        state1.lattice_part['cache'] = a ; state2.lattice_part['cache'] = b
        a.access(43) ; a.access(42) ; b.access(42)
        c = a.must_join(b)
        state2.lattice_part['cache'] = c
        self.assertTrue(state1 <= state2)
        self.assertFalse(state1 == state2)

    def test_abstract_cache_model(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_lattice_cache.xml'))
        nta = pyuppaal.from_xml(file)
        
        succgen = SuccessorGenerator(nta)
        goalchecker = GoalChecker(succgen, '')
        pwlist = LatticeCoverPWList()

        init_state = succgen.get_initialstate()

        #can we make a copy of the state?
        a = init_state.copy()

        succs = list(succgen.successors(init_state))
        self.assertEqual(len(succs), 1)
        succ1 = succs[0]
        #print succ1, succ1['cache']

        succs2 = list(succgen.successors(succs[0]))
        self.assertEqual(len(succs2), 1)
        succ2 = succs2[0]
        #print succ2, succ2['cache']

    def test_union_bitvector(self):
        a = UnionBitVector(8) ; b = UnionBitVector(8)
        
        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        b[3] = True
        self.assertTrue(a <= b)
        self.assertFalse(a == b)
        self.assertEqual(a.num1s(), 0)
        self.assertEqual(b.num1s(), 1)

        a[3] = True
        self.assertTrue(a <= b)
        self.assertTrue(a == b)
        self.assertEqual(a.num1s(), 1)
        self.assertEqual(b.num1s(), 1)

        a = UnionBitVector(8) ; b = UnionBitVector(8)

        a[3] = True
        b[5] = True
        self.assertEqual(a.num1s(), 1)
        self.assertEqual(b.num1s(), 1)

        c = a.join(b)
        self.assertFalse(c[0])
        self.assertFalse(c[1])
        self.assertFalse(c[2])
        self.assertTrue(c[3])
        self.assertFalse(c[4])
        self.assertTrue(c[5])
        self.assertFalse(c[6])
        self.assertFalse(c[7])
        self.assertEqual(c.num1s(), 2)

        self.assertTrue(a <= c)
        self.assertTrue(b <= c)
        self.assertFalse(a == c)
        self.assertFalse(b == c)

    def test_intersection_bitvector(self):
        a = IntersectionBitVector(8) ; b = IntersectionBitVector(8)
        
        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        a[3] = True
        self.assertTrue(a <= b)
        self.assertFalse(a == b)

        b[3] = True
        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        a = IntersectionBitVector(8) ; b = IntersectionBitVector(8)

        a[3] = True
        b[5] = True

        c = a.join(b)
        self.assertFalse(c[0])
        self.assertFalse(c[1])
        self.assertFalse(c[2])
        self.assertFalse(c[3])
        self.assertFalse(c[4])
        self.assertFalse(c[5])
        self.assertFalse(c[6])
        self.assertFalse(c[7])

        self.assertTrue(a <= c)
        self.assertTrue(b <= c)
        self.assertFalse(a == c)
        self.assertFalse(b == c)

        a = IntersectionBitVector(8) ; b = IntersectionBitVector(8)
        a[3] = True ; b[3] = True
        a[5] = True ; b[5] = True
        c = a.join(b)
        self.assertFalse(c[0])
        self.assertFalse(c[1])
        self.assertFalse(c[2])
        self.assertTrue(c[3])
        self.assertFalse(c[4])
        self.assertTrue(c[5])
        self.assertFalse(c[6])
        self.assertFalse(c[7])

    def test_lattice_vector(self):
        a = LatticeVector(UnionBitVector(8), 8)
        b = LatticeVector(UnionBitVector(8), 8)

        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        b[3][5] = True
        self.assertTrue(a <= b)
        self.assertFalse(a == b)

        a[5][3] = True
        self.assertFalse(a <= b)
        self.assertFalse(a == b)

        c = a.join(b)
        self.assertTrue(a <= c)
        self.assertTrue(b <= c)
        self.assertFalse(a == c)
        self.assertFalse(b == c)

        #test 3 dimensional
        a = LatticeVector(LatticeVector(UnionBitVector(8), 8), 8)
        b = LatticeVector(LatticeVector(UnionBitVector(8), 8), 8)
        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        b[3][3][3] = True
        self.assertTrue(a <= b)
        self.assertFalse(a == b)

        a[5][5][5] = True
        self.assertFalse(a <= b)
        self.assertFalse(a == b)

        c = a.join(b)
        self.assertTrue(a <= c)
        self.assertTrue(b <= c)
        self.assertFalse(a == c)
        self.assertFalse(b == c)

    def test_lattice_cover_bitvector(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_lattice_cover.xml'))
        nta = pyuppaal.from_xml(file)
        
        succgen = SuccessorGenerator(nta)
        goalchecker = GoalChecker(succgen, '')

        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertEqual(7, len(pwlist.passed_waiting))

        #check that lattice inclusion check succeeds manually
        succs = list(succgen.successors(init_state))
        self.assertEqual(2, len(succs))
        succ1 = succs[0] ; succ2 = succs[1]
        succ1n = list(succgen.successors(succ1))[0]
        succ2n = list(succgen.successors(succ2))[0]

        self.assertTrue(succ1n.discrete_eq(succ2n))
        self.assertEqual(hash(succ1n), hash(succ2n))
        self.assertFalse(succ1n == succ2n)
        #print succ1n, succ2n
        self.assertTrue(succ1n <= succ2n or succ2n <= succ1n)

        pwlist = LatticeCoverPWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        #print pwlist.passed_waiting
        self.assertEqual(5, len(pwlist.passed_waiting))

    def test_maxinteger_lattice(self):
        a = MaxInteger() ; b = MaxInteger();

        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        b += 5
        self.assertTrue(a <= b)
        self.assertFalse(a == b)

        a += 7
        self.assertTrue(b <= a)
        self.assertFalse(a == b)

        b += 2
        self.assertTrue(a <= b)
        self.assertTrue(a == b)

        a = MaxInteger() ; b = MaxInteger();
        a += 7
        b += 5
        c = a.join(b)
        self.assertEqual(int(c), 7)
        self.assertTrue(a <= c)
        self.assertTrue(b <= c)





if __name__ == '__main__':
    verbose_level = sys.argv.count('-v')
    logging.basicConfig( stream=sys.stderr, 
            level={0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbose_level] )

    unittest.main() 

# vim:ts=4:sw=4:expandtab
