#!/usr/bin/python
import sys
import os
import unittest
import logging

import pyuppaal
from opaal.model_parsers.pyuppaal import SuccessorGenerator
from opaal.model_parsers.pyuppaal.simplify import SimplifyModel

class TestPYUPPAALSuccGen(unittest.TestCase):

    def test_invariant(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_invariant.xml'))
        nta = pyuppaal.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 0)


    def test_simplify(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'master-election-simple.xml'))
        nta = pyuppaal.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        simple_nta = SimplifyModel(nta).simplify()
        
        #Test that templates have been expanded
        self.assertEqual(len(simple_nta.templates), 7)

        #test select has been expanded
        temp0 = [t for t in simple_nta.templates if t.name == 'Node_0'][0]
        self.assertEqual(len(temp0.transitions), 9)

    def test_simple_sync(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_simple_sync.xml'))
        nta = pyuppaal.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)
        
        succs2 = list(s.successors(succs[0]))
        self.assertEqual(len(succs2), 0)


    def test_array_init(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_array_init.xml'))
        nta = pyuppaal.from_xml(file)
        
        s = SuccessorGenerator(nta)

        self.assertEqual(s.variables, [('test', 'int', [2, 3]), ('test2', 'int', [2, 3, 4])])

        initial_state = s.get_initialstate()
        self.assertEqual(initial_state['test'], [[0, 0, 0], [0, 0, 0]])
        self.assertEqual(initial_state['test2'], [[[0]*4]*3]*2)

        initial_state['test'][0][0] = 42
        self.assertEqual(initial_state['test'], [[42, 0, 0], [0, 0, 0]])

    def test_simplify_channel_array(self):
        file = open(os.path.join(os.path.dirname(sys.argv[0]), 'test_channel_array.xml'))
        nta = pyuppaal.from_xml(file)

        simple_nta = SimplifyModel(nta).simplify()
        #test select has been expanded
        temp0 = [t for t in simple_nta.templates if t.name == 'Template_0'][0]
        self.assertEqual(len(temp0.transitions), 1)
        trans0 = temp0.transitions[0]
        self.assertEqual(trans0.synchronisation.value, 'mychan[0]!')

        #test synchronisation on channel array
        s = SuccessorGenerator(simple_nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 3)


if __name__ == '__main__':
    verbose_level = sys.argv.count('-v')
    logging.basicConfig( stream=sys.stderr, 
            level={0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbose_level] )

    unittest.main() 
